import requests
import pandas as pd
from urllib.parse import quote
from .utils import data_dict_to_df, add_list_to_dict, subtabtree


class StatBankClient:
    """Client that connects to the Databank API of Statistics Denmark.

    Attributes
    ----------
    lang : str, {'da', 'en'} default is 'da'
        Note: Variable key codes to be passed when requesting specific data
        must be in Danish. Please see data() method docstring for details. 
        As long as Danish key codes are provided as input, one can use the 
        Class with English language settings.
    """

    def __init__(self, lang='da'):
        self._lang = lang
        self.session = requests.session()
        self._base_url = 'https://api.statbank.dk/v1/'

    @property
    def lang(self):
        return self._lang

    @lang.setter
    def lang(self, value):
        if value not in ['en', 'da']:
            raise ValueError(
                'Language can only accept either "en" or "da" as values.')
        self._lang = value

    def base_request(self, cat, params):
        """Submits all POST request to API and returns response."""
        params.update({'lang': self.lang})
        try:
            resp = self.session.post(self._base_url+quote(cat), json=params)
            if resp.status_code == 200:
                return resp.json()
            else:
                response_message = resp.json()['message']
                print(response_message)
        except Exception as err:
            print(err)

    def subjects(self, subjects=None, include_tables=False, recursive=False, as_tree=True):
        """Retrieves the basic subject(s) information for which tables exist in
        the StatBank database.

        Parameters
        ----------
        subjects : list, default is None
            list of subject id strings
        include_tables : bool, default is False
            if True, returns list of table information attached to subject id.
        recursive : bool, default is False
            if True, subtopics/tables will be retrieved all the way down the
            hierarchy.
        as_tree : bool, default is True
            prints a tree structure of the subjects heirarchy.
            Also prints tables if include_tables is True.
            Each item is a tuple with (subject/table id, description/text)
            Note: subject id's are numeric characters whereas table id's
            are alphanumeric.
            If False, returns a list of dictionaries.

        Returns
        -------
        list of dicts, or prints a tree structure.

        Examples
        --------
        >>> sbc.subjects(['19'], recursive=True)
                 |--('19', 'Other')
                 |       |--('2473', 'Statistisk Årbog')
                 |       |--('2472', 'Statistisk Tiårsoversigt')
                 |       |--('2475', 'Danmark i Tal')
                 |       |--('2474', 'Nordic Statistical Yearbook')
                 |       |--('2483', 'Publications')
                 |       |       |--('10392', 'Generel publications')
                 |       |       |--('10393', 'Methodology')
                 |       |       |--('10394', 'Nomenclatures and classifications')
                 |       |--('2482', 'Municipal service indicators')
                 |       |       |--('10376', 'Elderly care')
                 |       |       |--('10378', 'Disadvantaged children and young people')
                 |       |       |--('10377', 'Health')ss
        """
        cat = 'subjects'
        params = dict(includeTables=include_tables, recursive=recursive)
        add_list_to_dict(params, subjects=subjects)
        resp = self.base_request(cat, params)
        if not as_tree:
            return resp
        else:
            if len(resp) > 0:
                for d in resp:
                    for g in subtabtree(d):
                        print(g)

    def tables(self, subjects=None, past_days=None, include_inactive=False, as_df=True):
        """Retrieves the complete list of tables present currently in the
        Statbank database together with relevant metadata.

        Parameters
        ----------
        subjects : list, default is None
            If provided, returns tables attached to subject id.
        past_days : int, default is None
            number of days from present time for which updated tables are
            available.
        include_inactive : bool, default is False
            if True, includes tables that are no longer active.
        as_df : bool, default is True
            If true, returns a pandas dataframe, otherwise returns a list of
            dictionaries.

        Returns
        -------
        pd.DataFrame or list of dicts

        Examples
        --------
        >>> df = sbc.tables()
        >>> df.iloc[0]
        id                                                  FOLK1A
        text            Population at the first day of the quarter
        unit                                                number
        updated                                2020-02-11T08:00:00
        firstPeriod                                         2008Q1
        latestPeriod                                        2020Q1
        active                                                True
        variables         [region, sex, age, marital status, time]
        Name: 0, dtype: object

        >>> df.shape
        (2116, 8)
        """
        cat = 'tables'
        params = dict(pastdays=past_days, includeinactive=include_inactive)
        add_list_to_dict(params, subjects=subjects)
        rjson = self.base_request(cat, params)
        if as_df:
            return pd.DataFrame(rjson)
        else:
            return rjson

    def tableinfo(self, table_id, variables_df=False):
        """Retrieves table specific information from the StatBank database.

        Parameters
        ----------
        table_id : str
        variables_df : bool, default is False
            If true, returns the variable names and possible values as a
            pandas dataframe, otherwise returns a dictionary including
            additional metadata.

        Returns
        -------
        dict or pd.DataFrame

        Examples
        --------
        >>> vdf = sbc.tableinfo(table_id='bef5', variables_df=True)
        >>> vdf.head()
          id     text variable
        0  M     Mænd      KØN
        1  K  Kvinder      KØN
        0  0     0 år    ALDER
        1  1     1 år    ALDER
        2  2     2 år    ALDER
        """
        cat = 'tableinfo'
        params = dict(table=table_id)
        rjson = self.base_request(cat, params)
        if variables_df:
            var_df = pd.DataFrame()
            varlist = rjson['variables']
            for d in varlist:
                df = pd.DataFrame(d['values'])
                if params['lang'] == 'en':
                    df['variable'] = d['text']
                else:
                    df['variable'] = d['id']
                var_df = var_df.append(df)
            return var_df
        else:
            return rjson

    def data(self, table_id, as_df=True, variables=None, **kwargs):
        """Retrieves the data for a specific table from the StatBank
        database.

        Parameters
        ----------
        table : str,
        as_df : bool, default is True
            If true, returns the essential data as a pandas dataframe.
            Otherwise returns a dictionary including additional metadata.
        variables : list, default is None
            List of dictionaries with variable parameters.
            Use make_variables_dict() method to generate these dictionaries
            in the prescribed format.
        optional kwargs

        Returns
        -------
        Single or multi-indexed pd.DataFrame or dict

        Examples
        --------
        >>> sbc.data(table_id='folk1a')
                Folketal den 1. i kvartalet efter Indhold og tid
            0                                           5822763

        >>> kon = sbc.variable_dict(code='KØN', values=['M', 'K'])
        >>> tid = sbc.variable_dict(code='Tid', values=['2018'])
        >>> df = sbc.data(table_id='bef5', variables=[tid, kon])
        >>> df
                    Population 1. January by sex, Indhold and time
        køn   tid
        Men   2018                                         2876473
        Women 2018                                         2904717
        """
        cat = 'data'
        params = dict(format='JSONSTAT', table=table_id)
        add_list_to_dict(params, variables=variables)
        params.update({k: v for k, v in kwargs.items() if k})
        codes = [d['code'].lower() for d in variables] if variables else []
        rjson = self.base_request(cat, params)
        if as_df:
            ddict = rjson['dataset']
            return data_dict_to_df(ddict, codes)
        else:
            return rjson

    @staticmethod
    def variable_dict(code, values, **kw):
        """Utility method to generate dictionary for a specific
        variable to select data from a table.
        ** Code and values must be in Danish. **

        Takes the generic form
        {'code': code, 'values': values}

        May contain optional key-value item such as
        {'placement': 'stub'}

        These dictionaries may be included as values to a
        dictionary with 'variables' key which may then be
        passed as kwarg to the data() method.

        Examples
        --------
        >>> tid = sbc.variable_dict('Tid', ['2018', '2019'])
        >>> tid
        {'code': 'Tid', 'values': ['2018', '2019']}
        """
        var_dict = {'code': [], 'values': []}
        var_dict['code'] = code
        if isinstance(values, list):
            var_dict['values'].extend(values)
        else:
            var_dict['values'].append(values)
        if kw:
            var_dict.update(kw)
        return var_dict



mock_sub_resp_default = [{'id': '02',
                          'description': 'Population and elections',
                          'active': True,
                          'hasSubjects': True,
                          'subjects': []},
                         {'id': '05',
                          'description': 'Living conditions',
                          'active': True,
                          'hasSubjects': True,
                          'subjects': []},
                         ]

mock_sub_resp_2401 = [{'id': '2401',
                       'description': 'Population and population projections',
                       'active': True,
                       'hasSubjects': True,
                       'subjects': [{'id': '10021',
                                     'description': 'Population in Denmark',
                                     'active': True,
                                     'hasSubjects': False,
                                     'subjects': []},
                                    {'id': '10022',
                                     'description': 'Population projections',
                                     'active': True,
                                     'hasSubjects': False,
                                     'subjects': []}]}]


mock_tables_resp = [{'id': 'FOLK1A',
                     'text': 'Population at the first day of the quarter',
                     'unit': 'number',
                     'updated': '2020-02-11T08:00:00',
                     'firstPeriod': '2008Q1',
                     'latestPeriod': '2020Q1',
                     'active': True,
                     'variables': ['region', 'sex', 'age', 'marital status', 'time']},
                    {'id': 'FOLK1B',
                     'text': 'Population at the first day of the quarter',
                     'unit': 'number',
                     'updated': '2020-02-11T08:00:00',
                     'firstPeriod': '2008Q1',
                     'latestPeriod': '2020Q1',
                     'active': True,
                     'variables': ['region', 'sex', 'age', 'citizenship', 'time']}]


mock_tableinfo_resp = {'id': 'FOLK1A',
                       'text': 'Population at the first day of the quarter',
                       'description': 'Population at the first day of the quarter by region, sex, age, marital status and time',
                       'unit': 'number',
                       'suppressedDataValue': '0',
                       'updated': '2020-02-11T08:00:00',
                       'active': True,
                       'footnote': None,
                       'variables': [{'id': 'KØN',
                                      'text': 'sex',
                                      'elimination': True,
                                      'time': False,
                                      'values': [{'id': 'TOT', 'text': 'Total'},
                                                 {'id': '1', 'text': 'Men'},
                                                 {'id': '2', 'text': 'Women'}]}]}


mock_tableinfo_variable_resp = {'id': ['000', '1', '32', 'U', '2008K4'],
                                'text': ['All  Denmark', 'Men', '32 years', 'Never married', '2008Q4'],
                                'variable': ['region', 'sex', 'age', 'marital status', 'time']
                                }


mock_data_resp = {'dataset': {'dimension': {'ContentsCode': {'label': 'Indhold',
                                                             'category': {'index': {'FOLK1A': 0},
                                                                          'label': {'FOLK1A': 'Population at the first day of the quarter'},
                                                                          'unit': {'FOLK1A': {'base': 'number', 'decimals': 0}}}},
                                            'Tid': {'label': 'time',
                                                    'category': {'index': {'2020K1': 0}, 'label': {'2020K1': '2020Q1'}}},
                                            'id': ['ContentsCode', 'Tid'],
                                            'size': [1, 1],
                                            'role': {'metric': ['ContentsCode'], 'time': ['Tid']}},
                              'label': 'Population at the first day of the quarter by Indhold and time',
                              'source': 'Statistics Denmark',
                              'updated': '2020-02-11T07:00:00Z',
                              'value': [5822763]}}


mock_data_resp_to_df = {
    ' Population at the first day of the quarter by Indhold and time': [5822763]}


mock_data_resp_with_vars = {'dimension': {'KØN': {'label': 'sex',
                                                  'category': {'index': {'M': 0, 'K': 1},
                                                               'label': {'M': 'Men', 'K': 'Women'}}},
                                          'FODLAND': {'label': 'country of birth',
                                                      'category': {'index': {'5101': 0, '5104': 1},
                                                                   'label': {'5101': 'Greenland', '5104': 'Finland'}}},
                                          'ContentsCode': {'label': 'Indhold',
                                                           'category': {'index': {'BEF5': 0},
                                                                        'label': {'BEF5': 'Population 1. January'},
                                                                        'unit': {'BEF5': {'base': 'number', 'decimals': 0}}}},
                                          'Tid': {'label': 'time',
                                                  'category': {'index': {'2018': 0, '2019': 1},
                                                               'label': {'2018': '2018', '2019': '2019'}}},
                                          'id': ['KØN', 'FODLAND', 'ContentsCode', 'Tid'],
                                          'size': [2, 2, 1, 2],
                                          'role': {'geo': ['FODLAND'], 'metric': ['ContentsCode'], 'time': ['Tid']}},
                            'label': 'Population 1. January by sex, country of birth, Indhold and time',
                            'source': 'Statistics Denmark',
                            'updated': '2020-02-11T07:00:00Z',
                            'value': [7016, 7095, 1266, 1263, 9454, 9471, 2766, 2787]}


mock_codes = ['køn', 'tid', 'fodland']



import pandas as pd


def data_dict_to_df(ddict, codes):
    """Augments a simple pandas dataframe to a multi-indexed dataframe from a
    list of key codes.
    """
    df = pd.DataFrame({ddict['label']: ddict['value']})
    dims = {k.lower(): v for k, v in ddict['dimension'].items()}
    keys = [k for k in dims.keys() if k in codes]
    vals = [list(dims[k]['category']['label'].values()) for k in keys]
    if len(vals) == len(keys) > 0:
        index = pd.MultiIndex.from_product(vals, names=keys)
        df.set_index(index, inplace=True)
    return df


def add_list_to_dict(d, **kwargs):
    """Adds keyword arguments whose values are lists to a given dictionary."""
    for k, v in kwargs.items():
        if v is not None:
            if isinstance(v, list):
                d.update({k: v})
            else:
                raise Exception(f'{k} must be a list.')


def subtabtree(d):
    """Generator to print tree structure of subjects and tables in db."""
    for k, v in d.items():
        if k == 'id':
            ival = d[k]
        if k == 'description' or k == 'text':
            dval = d[k]
            v = (ival, dval)
            yield f'\t |--{v}'
        if isinstance(v, list) and len(v) > 0:
            for i in v:
                if isinstance(i, dict):
                    for g in subtabtree(i):
                        yield f'\t |  {g}'



from os.path import abspath, dirname, join
import pkg_resources
import re
import setuptools


lpath = abspath(dirname(__file__))

with open(join(lpath, 'requirements.txt'), "r") as requirements:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements)
    ]


with open(join(lpath, "README.md"), "r") as fh:
    long_description = fh.read()

with open(join(lpath, "denstatbank/__init__.py"), "rt", encoding="utf8") as f:
    version = re.search(r'__version__\s=\s"(.*?)"', f.read()).group(1)

setuptools.setup(
    name="denstatbank",
    version=version,
    author="Gopakumar Mohandas",
    author_email="gopakumarmohandas@gmail.com",
    description="A Python API wrapper to Statistics Denmark's DataBank API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gmohandas/denstatbank.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.6',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=install_requires
)



import pandas as pd
import pytest
from denstatbank.denstatbank import StatBankClient
from denstatbank.utils import data_dict_to_df, add_list_to_dict
from .mock_responses import (
    mock_sub_resp_default,
    mock_sub_resp_2401,
    mock_tables_resp,
    mock_tableinfo_resp,
    mock_tableinfo_variable_resp,
    mock_data_resp,
    mock_data_resp_to_df,
    mock_data_resp_with_vars,
    mock_codes
)


@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """Remove requests.sessions.Session.request for all tests."""
    monkeypatch.delattr("requests.sessions.Session.request")


@pytest.fixture
def client():
    client = StatBankClient()
    return client


def test_base_request(client, monkeypatch):
    def mock_base_request(self, *args, **kwargs):
        return mock_tableinfo_resp
    monkeypatch.setattr(StatBankClient, "base_request", mock_base_request)
    r = client.base_request('data', lang='en')
    assert r == mock_tableinfo_resp


def test_subjects(client, monkeypatch):
    def mock_subjects(self, subjects=None, include_tables=False, recursive=False):
        if subjects is None:
            return mock_sub_resp_default
    monkeypatch.setattr(StatBankClient, "subjects", mock_subjects)
    r = client.subjects()
    assert isinstance(r, list)
    d = r[0]
    assert isinstance(d, dict)
    assert 'id' in d.keys()
    assert 'description' in d.keys()


def test_subjects_returns_specified_subject(client, monkeypatch):
    def mock_subjects(self, subjects=None, include_tables=False, recursive=False):
        if subjects[0] == '2401':
            return mock_sub_resp_2401
    monkeypatch.setattr(StatBankClient, "subjects", mock_subjects)
    r = client.subjects(subjects=['2401'])
    assert isinstance(r, list)
    d = r[0]
    assert isinstance(d, dict)
    assert d['id'] == '2401'


def test_tables_returns_dict(client, monkeypatch):
    def mock_tables(self, subjects=None, past_days=None, include_inactive=False, as_df=True):
        return mock_tables_resp
    monkeypatch.setattr(StatBankClient, "tables", mock_tables)
    r = client.tables(as_df=False)
    assert isinstance(r, list)
    d = r[0]
    assert isinstance(d, dict)
    assert 'id' in d.keys()
    assert 'text' in d.keys()
    assert 'unit' in d.keys()
    assert 'updated' in d.keys()
    assert 'firstPeriod' in d.keys()
    assert 'latestPeriod' in d.keys()
    assert 'active' in d.keys()
    assert 'variables' in d.keys()


def test_tables_returns_df(client, monkeypatch):
    def mock_tables(self, subjects=None, past_days=None, include_inactive=False, as_df=True):
        return pd.DataFrame(mock_tables_resp)
    monkeypatch.setattr(StatBankClient, "tables", mock_tables)
    df = client.tables()
    assert isinstance(df, pd.DataFrame)
    assert 'id' in df.columns
    assert 'text' in df.columns
    assert 'unit' in df.columns
    assert 'updated' in df.columns
    assert 'firstPeriod' in df.columns
    assert 'latestPeriod' in df.columns
    assert 'active' in df.columns
    assert 'variables' in df.columns


def test_tableinfo_returns_dict(client, monkeypatch):
    def mock_tableinfo(self, table_id, variables_df=False):
        return mock_tableinfo_resp
    monkeypatch.setattr(StatBankClient, "tableinfo", mock_tableinfo)
    d = client.tableinfo('FOLK1A')
    assert isinstance(d, dict)
    assert d['id'] == 'FOLK1A'
    assert 'id' in d['variables'][0].keys()
    assert 'text' in d['variables'][0].keys()


def test_tableinfo_returns_variables_df(client, monkeypatch):
    def mock_tableinfo(self, table_id, variables_df):
        if variables_df:
            return pd.DataFrame(mock_tableinfo_variable_resp)
    monkeypatch.setattr(StatBankClient, "tableinfo", mock_tableinfo)
    df = client.tableinfo('FOLK1A', variables_df=True)
    assert isinstance(df, pd.DataFrame)
    print(df)
    assert 'id' in df.columns
    assert 'text' in df.columns
    assert 'variable' in df.columns
    assert len(df.columns.tolist()) == 3


def test_data_returns_dict(client, monkeypatch):
    def mock_data(self, table_id, as_df, variables=None, **kwargs):
        return mock_data_resp
    monkeypatch.setattr(StatBankClient, "data", mock_data)
    d = client.data(table_id='folk1a', as_df=False)
    assert isinstance(d, dict)
    assert 'dataset' in d.keys()
    dd = d['dataset']
    assert 'value' in dd.keys()
    assert isinstance(dd['value'], list)


def test_data_returns_df(client, monkeypatch):
    def mock_data(self, table_id, as_df=True, variables=None, **kwargs):
        return pd.DataFrame(mock_data_resp_to_df)
    monkeypatch.setattr(StatBankClient, "data", mock_data)
    d = client.data(table_id='folk1a')
    assert isinstance(d, pd.DataFrame)


def test_variables_dict(client):
    kon = client.variable_dict(code='køn', values=['M', 'K'])
    assert isinstance(kon, dict)
    assert 'code' in kon.keys()
    assert 'values' in kon.keys()
    assert kon['code'] == 'køn'
    assert isinstance(kon['values'], list)
    assert kon['values'] == ['M', 'K']

    tid = client.variable_dict(code='tid', values='2018')
    assert isinstance(tid, dict)
    assert 'code' in tid.keys()
    assert 'values' in tid.keys()
    assert tid['code'] == 'tid'
    assert isinstance(tid['values'], list)
    assert tid['values'] == ['2018']


def test_data_dict_to_df():
    df = data_dict_to_df(mock_data_resp_with_vars, mock_codes)
    assert isinstance(df, pd.DataFrame)
    assert isinstance(df.index, pd.MultiIndex)
    assert df.shape == (8, 1)


def test_add_list_to_dict():
    params = {'lang': 'en'}
    add_list_to_dict(params, subjects=['02'])
    assert 'subjects' in params.keys()
    assert isinstance(params['subjects'], list)
    assert params['subjects'] == ['02']

    with pytest.raises(Exception) as e:
        assert add_list_to_dict(params, subjects='03')
    assert str(e.value) == 'subjects must be a list.'
